#!/usr/bin/perl
#
# $Id: webfm.cgi,v 1.10 2000/10/03 15:25:45 nrjryka Exp $
#
# NAME
#	webfm.cgi -- File manager with HTML interface
#
# DESCRIPTION
#	This CGI script is a file manager that runs on the web server.
#	The main purpose is to maintain the web page from anywhere
#	you accesses the Internet. It can upload, download, rm,
#	chmod, mkdir and rmdir through any web browser even if it is 
#	located behind the firewall.
#
# HISTORY
#	24/Apr/02 Ver0.60  "edit" was added
#	03/Oct/00 Ver0.51  smart dir selection for upload & mkdir
#	03/Oct/00 Ver0.50  enable to handle multiple files
#	03/Oct/00 Ver0.44  temtative version with full functionality
#	03/Oct/00 Ver0.43  temtative version
#	02/Oct/00 Ver0.42  temtative version
#	01/Oct/00 Ver0.41  cosmetic change
#	01/Oct/00 Ver0.40  multiple files upload
#	29/Sep/00 Ver0.33  all cosmetic in print_webfm
#	29/Sep/00 Ver0.32  fix the wrong parent dir name
#	28/Sep/00 Ver0.31  drop ls function
#	28/Sep/00 Ver0.30  new HTML interface
#	27/Sep/00 Ver0.21  keep the file permission same when upload
#	26/Sep/00 Ver0.20  mkdir, rmdir and ls works
#	25/Sep/00 Ver0.12  download works
#	25/Sep/00 Ver0.11  get/post url encoded paramters are acceptable
#	12/Sep/00 Ver0.10  upload and delete works
#	11/Sep/00 Ver0.01  got to start
use strict;
use warnings;

########################################################################
###
### configure
###

my $TOPDIR			= '.';
my $JP				= 1;

my $DEFAULT_EOL			= "\015\012";	# 'dos' or 'unix'
my $DEFAULT_MODE		= 0600;
my $DEFAULT_DIRMODE		= 0700;
my $DEFAULT_TIMEOUT		= 5;
my $DEFAULT_JCODE		= 'sjis';	# euc, jis or sjis
my $DEFAULT_BLOCKSIZE		= 4096;

my $HEAD_BANNER			= undef;
my $TAIL_BANNER			= <<__XREA_AD__;
<hr>
<SCRIPT LANGUAGE="JAVASCRIPT"><!-- 
window.open("http://ad.xrea.com/ad_pop.fcg?site=hehe.s9.xrea.com","XREAADPOP","width=500,height=70"); 
//--></SCRIPT> 
<NOSCRIPT>
<A HREF="http://ad.xrea.com/ad_click.fcg?site=hehe.s9.xrea.com" TARGET="_BLANK" ><IMG SRC="http://ad.xrea.com/ad_img.fcg?site=hehe.s9.xrea.com" HEIGHT="60" WIDTH="468" BORDER="0"></A>
</NOSCRIPT>
<!--nobanner--> 
__XREA_AD__

########################################################################
###
### debug
###

my $DEBUG;			# init in BEGIN
my $DEBUG_SAVE			= 0;
my $DEBUG_ENV_FILE = "DEBUG_ENV_FILE";
my $DEBUG_STDIN_FILE = "DEBUG_STDIN_FILE";

########################################################################
###
### const
###

my $CR = "\015";			# carriage return
my $LF = "\012";			# line feed or unix end-of-line
my $CRLF = "\015\012";			# dos (and html) end-of-line
my $DOS_EOL = $CRLF;
my $UNIX_EOL = $LF;

my %ESCCHAR = (				# escape character in HTML
	'<'	=> '&lt;',
	'>'	=> '&gt;',
	'&'	=> '&amp;',
	'"'	=> '&quot;',
	' '	=> '&nbsp;'
	);
my @MONTH = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
             'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');

########################################################################
###
### valiable
###

my %param = ();				# CGI parameters
					# current time
my ( $lsec, $lmin, $lhour, $lmday,
     $lmon, $lyear, $lwday, $lyday, $lisdst) = localtime();

my %upfiles = ();			# temporary uploaded files

########################################################################
###
### BEGIN block for valiable init
###

BEGIN {
	$DEBUG = 0;
}

############################################################################
###
### package tmpfile
###

package tmpfile;

our @files = ();
our $i = 0;

sub name() {
	my $file;

	do {
		++$i;
		$file = "__tmpfile$i";
	}
	while ( -e $file );
	push(@files, $file);
	return($file);
}

sub clean() {
	while ( my $file = pop @files ) {
		unlink($file) if ( -e $file );
	}
}

sub END {
	clean();
};

1;

############################################################################
###
### package jp
###
package jp;

our $type;				# 0(disable) 1(jcode.pl) 2(Jcode.pm)
our $on;

sub BEGIN {
	@INC = ( ".", @INC );

	my $val = eval { require Jcode; }; 
	if ( $val ) {
		$type = 2;
	}
	else {
		$val = eval { require 'jcode.pl'; };
		if ( $val ) {
			$type = 1
		}
		else {
			$type = 0;
		}
	}
	$on = $type;
}

sub able() {
	return $on;
}

sub enable() {
	$on = ($type > 0);
	return $on;
}

sub disable() {
	$on = 0;
	return $on;
}

sub convert($;$) {
	my $str = shift;
	my $jcode = shift || $DEFAULT_JCODE;

	if ( $jcode !~ /euc|jis|sjis/i ) {
		return $str;
	}
	elsif ( $type == 1) {
		eval { return &jcode::to($jcode, $str); };
	}
	elsif ( $type == 2) {
		eval { return &Jcode::convert($str, $jcode); };
	}
	else {
		return $str;
	}
}

sub getcode($) {
	my $str = shift;
	my $code = undef;

	if ( $type == 1) {
		eval { $code = &jcode::getcode($str); };
	}
	elsif ( $type == 2) {
		eval { $code = &Jcode::getcode($str); };
	}
	return $code;
}
1;

############################################################################
###
### package eol
###
package eol;

sub convert($;$) {
	my $str = shift;
	my $eol = shift || $DEFAULT_EOL;

	if ( $eol eq "\015\012" ) {
		$str =~ s/\015?\012$/\015\012/;
	}
	elsif ( $eol eq "\012" ) {
		$str =~ s/\015?\012$/\012/;
	}
	else {
		;
	}
	return $str;
}

sub geteol($) {
	my $str = shift;

	if ( $str =~ /\015\012$/ ) {
		return "\015\012";
	}
	elsif ( $str =~ /\012$/ ) {
		return "\012";
	}
	else {
		return undef;
	}
}

1;

package main;

########################################################################
###
### environment variables
###
my $script_name = $ENV{'SCRIPT_NAME'};
my $request_method = $ENV{'REQUEST_METHOD'};
my $content_type = $ENV{'CONTENT_TYPE'};
my $content_length = $ENV{'CONTENT_LENGTH'};
my $query_string = $ENV{'QUERY_STRING'};

########################################################################
###
### program start, end
###

sub BEGIN {

	if ( $DEBUG ) {
		print qq(Content-Type: text/plain\n\n);
	}
}

sub END {
	if ( $DEBUG_SAVE ) {
		close(DBGOUT);
	}
}

START: {
	if ( $DEBUG_SAVE ) {
		open(DBGOUT, "> $DEBUG_ENV_FILE");
		printf DBGOUT "SCRIPT_NAME=\'%s\' export SCRIPT_NAME\n",
			$script_name;
		printf DBGOUT "REQUEST_METHOD=\'%s\' export REQUEST_METHOD\n",
			$request_method;
		printf DBGOUT "CONTENT_TYPE=\'%s\' export CONTENT_TYPE\n",
			$content_type;
		printf DBGOUT "CONTENT_LENGTH=\'%s\' export CONTENT_LENGTH\n",
			$content_length;
		printf DBGOUT "QUERY_STRING=\'%s\' export QUERY_STRING\n",
			$query_string;
		close(DBGOUT);
	}
	if ( $DEBUG_SAVE ) {
		open(DBGOUT, "> $DEBUG_STDIN_FILE");
		binmode(DBGOUT);

		## close(DBGOUT) in sub END
	}
}

########################################################################
###
### string converter
###

sub escape_html($) {
	my $str = shift;

	$str =~ s/([\<\>\&\"])/$ESCCHAR{$1}/g;
	return $str;
}

sub escape_url($) {
	my $str = shift;

	$str =~ s/([^\.\*\-_a-zA-Z0-9 ])/sprintf("%%%02lX",unpack("C",$1))/eg;
	$str =~ s/ /+/g;
	
	return $str;
}

sub unescape_url($) {
	my $str = shift;

	$str =~ tr/+/ /;
	$str =~ s/%([0-9a-fA-F][0-9a-fA-F])/pack("C", hex($1))/eg;
	
	return $str;
}

############################################################################
###
### file manager engine
###

#
# decode multipart/form_data
#
sub decode_formdata($) {
	my $formdata = shift;
	my ( $header, $body ) = split("${CRLF}${CRLF}", $formdata, 2);
	my $val;

	for (split("${CRLF}", $header)) {
		if ((/^Content-Disposition: ([^;]*); name="([^;]*)"; filename="([^\"]*)"/i) ||
		    (/^Content-Disposition: ([^;]*); name="upload::(file\d*)::([^\"]*)"/i)) {
			$val = tmpfile::name();
			if ( $3 ne "" ) {
				$param{$2} = $3;
				$upfiles{$3} = $val;
			}
			open(OUT, "> $val");
			binmode(OUT);
			syswrite(OUT, $body, length($body));
			close(OUT);
		}
		elsif (/^Content-Disposition: ([^;]*); name="([^;]*)"/i) {
			$param{$2} = $body;
		}
	}
}
		
#
# decode the received data from STDIN on the fly
#
sub decode_stdin() {
	my $val;
	my $rcvlen = 0;
	my $blocksize = $DEFAULT_BLOCKSIZE;
	my $block = "";
	my $delimiter = "";
	my $delimiter_eof = "";
	my $margin = 0;
	my ( $rin, $rout, $ein, $eout );
	my $stdin_fileno = fileno(STDIN);
	my $buf;

	$rin = '';
	vec( $rin, $stdin_fileno, 1 ) = 1;
	$ein = $rin;

	binmode(STDIN);

LOOP:
	while (1) {
		my $readlen = $blocksize - length($block);

		# read stdin
		#
		while (($readlen) && ($rcvlen < $content_length)) {

			select($rout=$rin, undef, $eout=$ein, $DEFAULT_TIMEOUT);
			if ((vec($eout, $stdin_fileno, 1)) ||
			    ( ! vec($rout, $stdin_fileno, 1))) {
				last LOOP;
			}

			$val = sysread(STDIN, $buf, $readlen);
			if ( $DEBUG_SAVE ) {
				syswrite(DBGOUT, $buf, $val);
			}

			$readlen -= $val;
			$rcvlen += $val;
			$block .= $buf;
		}

		# delimiter if not defined yet
		#
		if (($delimiter eq "") && ($block =~ /^(.*?)${CRLF}/s)) {
			$delimiter = "$1${CRLF}";
			$delimiter_eof = "$1--${CRLF}";
			$margin = length("${CRLF}${delimiter_eof}");
		}

		# -----------------------delimiter
		# ...
		# -----------------------delimiter
		#
		if ($block =~ /^${delimiter}(.*?)${CRLF}${delimiter}/s) {
			&decode_formdata($1);
			$block = substr($block, length($delimiter) +
			                        length($1) + 2);
		}
		# -----------------------delimiter
		# ...
		# -----------------------delimiter--
		#
		elsif ($block =~ /^${delimiter}(.*?)${CRLF}${delimiter_eof}$/s) {
			&decode_formdata($1);
			last LOOP;
		}
		# -----------------------delimiter
		# Content-Disposition: ...
		#
		elsif (($block =~ /^${delimiter}Content-Disposition: ([^;]*); name="([^;]*)"; filename="([^\"]*)".*?${CRLF}${CRLF}(.+)/si) ||
		       ($block =~ /^${delimiter}Content-Disposition: ([^;]*); name="upload::(file\d*)::([^\"]*)".*?${CRLF}${CRLF}(.+)/si)) {
			$val = tmpfile::name();
			if ( $3 ne "" ) {
				$param{$2} = $3;
				$upfiles{$3} = $val;
			}
			open(OUT, "> $val");
			binmode(OUT);
			if (length($4) > $margin) {
				syswrite(OUT, $4, length($4) - $margin);
				$block = substr($4, length($4) - $margin);
			}
		}
		# ...
		# -----------------------delimiter
		#
		elsif ($block =~ /^(.*?)${CRLF}${delimiter}/s) {
			syswrite(OUT, $1, length($1));
			close(OUT);
			$block = substr($block, length($1) + 2);
		}
		# ...
		# -----------------------delimiter--
		#
		elsif ($block =~ /^(.*?)${CRLF}${delimiter_eof}$/s) {
			syswrite(OUT, $1, length($1));
			close(OUT);
			last LOOP;
		}
		# $block > length("-----------------------delimiter--")
		#
		elsif (length($block) > $margin) {
			syswrite(OUT, $block, length($block) - $margin);
			$block = substr($block, length($block) - $margin);
		}
		# few data in $block
		#
		else {
			;			# read more from stdin
		}
	}
}

#
# decode www-urlencoded string
#
sub decode_string($) {
	my @pair = split(/&/, shift);
	my $str;

	foreach $str (@pair) {
		my ( $name, $val ) = split(/=/, $str);
		$name = &unescape_url($name);
		$val = &unescape_url($val);
		$param{$name} = $val;
	}
}

############################################################################
###
### file manager engine
###

#
# execute download
#
sub exec_download($) {
	my $file = shift;
	my ( $val, $buf );

	return if ( ! -r $file );

	print qq(Content-Type: application/octet-stream\n);
	print qq(\n);

	open(IN, $file);
	binmode(IN);
	binmode(STDOUT);

	while ($val = read(IN, $buf, $DEFAULT_BLOCKSIZE)) {
		print $buf;
	}
	close(IN);
}

sub exec_convert($$;$$) {
	my ( $ofile, $nfile, $dos, $code ) = @_;

	$dos = 0 if ( ! defined $dos );
	$code = 0 if ( ! defined $code );

	if (( $dos == 0 ) && ( $code == 0 )) {
		rename($ofile, $nfile);
	}
	else {
		open(IN, $ofile);
		open(OUT, "> $nfile");
		while (my $buf = <IN>) {
			if ( jp::able ) {
				if ( $code == 1 ) {
					$buf = &jp::convert($buf, 'euc');
				}
				elsif ( $code == 2 ) {
					$buf = &jp::convert($buf, 'jis');
				}
				elsif ( $code == 3 ) {
					$buf = &jp::convert($buf, 'sjis');
				}
			}
			
			if ( $dos == 1 ) {
				print OUT eol::convert($buf, $DOS_EOL);
			}
			else {
				print OUT eol::convert($buf, $UNIX_EOL);
			}
		}
		close(OUT);
		close(IN);
	}
}

#
# execute command
#
sub exec_cmd() {
	if ( ! defined $param{'cmd'} ) {
		foreach (keys %param) {
			if ( /^cmd_(.*?)(?:\.[xy])?$/ ) {
				$param{'cmd'} = $1;
				last;
			}
		}
	}
	
	if ( ! defined $param{'cmd'} ) {
		&print_webfm();
	}
	elsif ($param{'cmd'} eq "exec_upload") {
		foreach (keys %param) {
			my ( $file, $mode );
			my $i;

			next if ( ! /^file(\d*)$/i);
			$i = $1;
			$param{$_} =~ /([^\\\/]*)$/i;
			$file = "${TOPDIR}/$param{'dir'}/$1";
			$mode = $DEFAULT_MODE;
			if ( -f $file ) {
				my @val = stat $file;
				$mode = $val[2];
			}
			if (( exists $param{"eol$i"} ) ||
			    ( exists $param{"jcode$i"} )) {
				exec_convert($upfiles{$param{$_}},
					$file, $param{"eol$i"},
					$param{"jcode$i"});
			}
			else {
				rename($upfiles{$param{$_}}, $file);
			}
			chmod($mode, $file);
		}
		&print_webfm($param{'dir'});
	}
	elsif ($param{'cmd'} eq "download") {
		foreach (keys %param) {
			my $file = "${TOPDIR}/$param{$_}";
			next if ( ! /^file\d*$/i);
			next if ( ! -f $file );

			exec_download($file);
			last;
		}
	}
	elsif ($param{'cmd'} eq "delete") {
		foreach (keys %param) {
			my $file = "${TOPDIR}/$param{$_}";
			next if ( ! /^file\d*$/i);
			next if ( ! -e $file );

			if ( -f $file ) {
				unlink($file);
			}
			elsif ( -d $file ) {
				rmdir($file);
			}
		}
		&print_webfm($param{'curdir'});
	}
	elsif ($param{'cmd'} eq "exec_chmod") {
		foreach (keys %param) {
			my $file = "${TOPDIR}/$param{$_}";
			next if ( ! /^file\d*$/i);
			next if ( ! -e $file );

			chmod(oct $param{'mod'}, ${file});
		}
		&print_webfm($param{'curdir'});
	}
	elsif ($param{'cmd'} eq "exec_rename") {
		foreach (keys %param) {
			my $file = "${TOPDIR}/$param{$_}";
			next if ( ! /^file(\d*)$/i);
			my $val = $1;
			next if ( ! -e $file );

			my $newname = sprintf("%s/%s",
				&parentdir($file),
				$param{"newname${val}"});
			rename($file, $newname);
		}
		&print_webfm($param{'curdir'});
	}
	elsif ($param{'cmd'} eq "exec_mkdir") {
		foreach (keys %param) {
			my $newdir = "${TOPDIR}/$param{'dir'}/$param{$_}";

			next if ( ! /^dirname\d*$/i);
			next if ( -e ${newdir} );
	
			mkdir(${newdir}, $DEFAULT_DIRMODE);
		}
		&print_webfm($param{'dir'});
	}
	elsif ($param{'cmd'} eq "exec_move") {
		foreach (keys %param) {
			next if ( ! /^file\d*$/i);
			my $oldname = "${TOPDIR}/$param{$_}";
			next if ( ! -e ${oldname} );
			$oldname =~ /([^\\\/]*)$/i;
			my $newname = "${TOPDIR}/$param{'destdir'}/$1";

			rename($oldname, $newname);
		}
		&print_webfm($param{'destdir'});
	}
	elsif ($param{'cmd'} eq "edit") {
		&print_edit(\%param);
	}
	elsif ($param{'cmd'} eq "rename") {
		&print_rename(\%param);
	}
	elsif ($param{'cmd'} eq "webfm") {
		&print_webfm("$param{'file'}");
	}
	elsif ($param{'cmd'} eq "upload") {
		&print_upload(\%param);
	}
	elsif ($param{'cmd'} eq "move") {
		&print_move(\%param);
	}
	elsif ($param{'cmd'} eq "mkdir") {
		&print_mkdir(\%param);
	}
	elsif ($param{'cmd'} eq "chmod") {
		&print_chmod(\%param);
	}
	else {
		&print_webfm();
	}
}
		
############################################################################
### FILE MANAGER HTML INTERFACE
############################################################################
#
# HTML template
#
sub print_cgi_header {
	print qq(Content-Type: text/html${CRLF});
	print qq(${CRLF});
	print qq(<html>${CRLF});
	print qq(<head>${CRLF});
	print qq(<title> Web File Manager </title>${CRLF});
	print qq(</head>${CRLF});
	print qq(<body>${CRLF});
	if ( defined $HEAD_BANNER ) {
		print $HEAD_BANNER;
	}
}

sub print_cgi_trailer {
	if ( defined $TAIL_BANNER ) {
		print $TAIL_BANNER;
	}
	print qq(</body>${CRLF});
	print qq(</html>);
}

sub make_url {
	my ( %data ) = @_;
	my $str = undef;

	while (my ($name, $val) = each(%data)) {
		$str .= &escape_url($name) . '=' . &escape_url($val) . '&';
	}
	if ( defined $str ) {
		$str =~ s/&$//;
		return ${script_name} . '?' . $str;
	}
	else {
		return ${script_name};
	}
}
		
sub print_stat {
	my ( $filename ) = @_;
	my ( $file ) = ( "${TOPDIR}/${filename}" );

	return if ( ! -e $file );

	my ( $dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size,
	     $atime, $mtime, $ctime, $blksize, $blocks ) = stat $file;
	
	$filename =~ /([^\\\/]*)$/i;
	$filename = $1;
	
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) 
		= localtime($mtime);

	printf("%s", &mode_str($mode));
	print qq(</td><td align="right">);
	print qq($size);
	print qq(</td><td>);
	printf("%s", $MONTH[$mon]);
	print qq(</td><td align="right">);
	printf("%s", $mday);
	print qq(</td><td>);
	if ($lyear == $year) {
		printf("%02d:%02d", $hour, $min);
	}
	else {
		printf("%05d",  $year + 1900);
	}
}

sub print_parentdir {
	my ( $dirname ) = @_;
	my ( $parent, $parentdir_url );

	if (( defined $dirname ) &&
	    ( $dirname ne "") &&
	    ( $dirname !~ /^[\\\/]$/)) {
		$parent = &parentdir($dirname);
		$parentdir_url = ( length($parent) > 0 ) ?
				&make_url (
					'cmd' => 'webfm',
					'file' => $parent
					) :
				$script_name;
		print qq(<a href=\"${parentdir_url}\">Up to parent</a>);
	}
}

sub print_diropt {
	my $dir = shift || "";
	my $path = (${dir} ne "") ? "${TOPDIR}/${dir}" : ${TOPDIR};

	print qq(<option value="${dir}">/${dir}${CRLF});

	opendir(DIR, $path);
	foreach (sort readdir(DIR)) {
		next if (/^\.+$/);
		next if ( ! -d "${path}/$_" );
		print_diropt((${dir} ne "") ? "${dir}/$_" : $_);
	}
}

sub print_checked {
	my ( $par ) = @_;

	print qq(<table>${CRLF});
	foreach (keys %$par) {
		next if ( ! /^file\d*$/i );
		
		$par->{$_} =~ /([^\\\/]*)$/i;
		print qq(<tr><td>);
		print qq(<b>$1</b>);
		print qq(</td><td>);
		print &print_stat($par->{$_});
		print qq(</td></tr>);
	}
	print qq(</table>${CRLF});

	foreach (keys %$par) {
		next if ( ! /^file\d*$/i );
		print qq(<input type=hidden name="$_" value="$par->{$_}">);
	}
}

sub print_move {
	my ( $par ) = @_;

	print_cgi_header;
	print qq(<h2>Directory /$par->{'curdir'}</h2><br>${CRLF});
	print qq(<form method=post enctype="application/x-www-form-urlencoded" action="$script_name">);
	print qq(<input type=hidden name="cmd" value="exec_move">${CRLF});
	print qq(<input type=hidden name="curdir" value="$par->{'curdir'}">${CRLF});
	&print_checked($par);
	print qq(<select name="destdir">);
	&print_diropt();
	print qq(</select>);
	print qq(<input type=submit value="move">);
	print qq(<input type=reset value="reset">);
	print qq(</form>);
	print_cgi_trailer;
}

sub print_rename {
	my ( $par ) = @_;
	my $i;

	print_cgi_header;
	print qq(<h2>Directory /$par->{'curdir'}</h2><br>${CRLF});
	print qq(<form method=post enctype="application/x-www-form-urlencoded" action="$script_name">);
	print qq(<input type=hidden name="cmd" value="exec_rename">${CRLF});
	print qq(<input type=hidden name="curdir" value="$par->{'curdir'}">${CRLF});
	print qq(<table>${CRLF});
	foreach (keys %$par) {
		my $file = "${TOPDIR}/$par->{$_}";
		next if ( ! /^file(\d*)$/i);
		my $val = $1;
		next if ( ! -e $file );

		$file =~ /([^\\\/]*)$/i;
		print qq(<tr><td>${CRLF});
		print qq(<input type=hidden name="file${val}" value="${file}">);
		print qq(<input type=text name="newname${val}" value="$1" size=30>);
		print qq(</td><td>${CRLF});
		&print_stat($par->{$_});
		print qq(</td></tr>${CRLF});
	}
	print qq(</table>${CRLF});
	print qq(<input type=submit value="rename">${CRLF});
	print qq(<input type=reset value="reset">${CRLF});
	print qq(</form>${CRLF});
	print_cgi_trailer;
}

sub print_chmod {
	my ( $par ) = @_;

	print_cgi_header;
	print qq(<h2>Directory /$par->{'curdir'}</h2><br>${CRLF});

	print qq(<form method=post enctype="application/x-www-form-urlencoded" action="$script_name">);
	print qq(<input type=hidden name="cmd" value="exec_chmod">);
	print qq(<input type=hidden name="curdir" value="$par->{'curdir'}">${CRLF});

	&print_checked($par);

	print qq(<br>);
	print qq(Permission: <input type=text name="mod" size=3><br>${CRLF});
	print qq(<input type=submit value="chmod">);
	print qq(<input type=reset value="reset">);
	print qq(</form>);
	print_cgi_trailer;
}


sub print_mkdir {
	my ( $par ) = @_;
	my $dirname = $par->{'curdir'};

	foreach (keys %$par) {
		next if ( ! /^file(\d*)$/i);
		my $dir = "${TOPDIR}/$par->{$_}";
		next if ( ! -d $dir );
		$dirname = $par->{$_};
		last;
	}

	print_cgi_header;
	print qq(<h2>New Directories on /${dirname}</h2><br>${CRLF});
	print qq(<br>${CRLF});

	print qq(<form method=post enctype="application/x-www-form-urlencoded" action="$script_name">${CRLF});
	print qq(<input type=hidden name="cmd" value="exec_mkdir">${CRLF});
	print qq(<input type=hidden name="dir" value="$dirname">${CRLF});
	print qq(Dir1: <input type=text name="dirname1" size=30><br>${CRLF});
	print qq(Dir2: <input type=text name="dirname2" size=30><br>${CRLF});
	print qq(Dir3: <input type=text name="dirname3" size=30><br>${CRLF});
	print qq(Dir4: <input type=text name="dirname4" size=30><br>${CRLF});
	print qq(Dir5: <input type=text name="dirname5" size=30><br>${CRLF});
	print qq(Dir6: <input type=text name="dirname6" size=30><br>${CRLF});
	print qq(Dir7: <input type=text name="dirname7" size=30><br>${CRLF});
	print qq(Dir8: <input type=text name="dirname8" size=30><br>${CRLF});
	print qq(<br>${CRLF});
	print qq(<input type=submit value="mkdir">${CRLF});
	print qq(<input type=reset value="reset">${CRLF});
	print qq(</form>${CRLF});
	print_cgi_trailer;
}

sub print_upload {
	my ( $par ) = @_;
	my $dirname = $par->{'curdir'};

	foreach (keys %$par) {
		next if ( ! /^file(\d*)$/i);
		my $dir = "${TOPDIR}/$par->{$_}";
		next if ( ! -d $dir );
		$dirname = $par->{$_};
		last;
	}

	print_cgi_header;
	print qq(<h2>Upload to /${dirname}</h2><br>${CRLF});
	print qq(<br>${CRLF});

	print qq(<form method=post enctype="multipart/form-data" action="$script_name">${CRLF});
	print qq(<input type=hidden name="cmd" value="exec_upload">${CRLF});
	print qq(<input type=hidden name="dir" value="$dirname">${CRLF});
	print qq(File1: <input type=file name="file1" size=30><br>${CRLF});
	print qq(File2: <input type=file name="file2" size=30><br>${CRLF});
	print qq(File3: <input type=file name="file3" size=30><br>${CRLF});
	print qq(File4: <input type=file name="file4" size=30><br>${CRLF});
	print qq(File5: <input type=file name="file5" size=30><br>${CRLF});
	print qq(File6: <input type=file name="file6" size=30><br>${CRLF});
	print qq(File7: <input type=file name="file7" size=30><br>${CRLF});
	print qq(File8: <input type=file name="file8" size=30><br>${CRLF});
	print qq(<br>${CRLF});
	print qq(<input type=submit value="upload">${CRLF});
	print qq(<input type=reset value="reset">${CRLF});
	print qq(</form>${CRLF});
	print_cgi_trailer;
}

sub print_edit {
	my ( $par ) = @_;
	my $dirname = $par->{'curdir'};
	my ( $file, $filename, $i );

	foreach (keys %$par) {
		next if ( ! /^file(\d*)$/i);
		my $dir = "${TOPDIR}/$par->{$_}";
		next if ( ! -d $dir );
		$dirname = $par->{$_};
		last;
	}

	print_cgi_header;
	print qq(<h2>Edit /${dirname}</h2><br>${CRLF});
	print qq(<br>${CRLF});

	print qq(<form method=post enctype="multipart/form-data" action="$script_name">${CRLF});
	print qq(<input type=hidden name="cmd" value="exec_upload">${CRLF});
	print qq(<input type=hidden name="dir" value="$dirname">${CRLF});

	$i=0;
	foreach (keys %param) {
		my $buf;
		my ( $code, $eol ) = ( undef, undef );

		$file = "${TOPDIR}/$param{$_}";
		next if ( ! /^file\d*$/i);
		next if ( ! -f $file );
		next if ( ! -r $file );

		$file =~ /([^\\\/]*)$/i;
		$filename = $1;
		++$i;

		open(IN, $file);
		binmode(IN);
		binmode(STDOUT);

		print qq(${filename}<br>${CRLF});

		print qq(<textarea cols=80 rows=16 name="upload::file${i}::$file">${CRLF});
		while ($buf = <IN>) {
			if ( ! defined $eol ) {
				$eol = eol::geteol($buf);
			}
			if (( jp::able ) && ( ! defined $code )) {
				$code = jp::getcode($buf);
				$code = undef if (( defined $code ) && ( $code eq 'ascii' ));
			}

			$buf = jp::convert($buf, 'euc');
			$buf = escape_html($buf);
			$buf = eol::convert($buf, $CRLF);
			$buf = jp::convert($buf, $DEFAULT_JCODE);
			print $buf;
		}
		close(IN);
		print qq(</textarea>${CRLF});

		## print "dos or unix ?" radio button
		##
		printf qq(<input type=radio name="eol${i}" value=1 %s> dos${CRLF}), ((defined $eol) && ($eol eq "$CRLF")) ? "checked" : "";
		printf qq(<input type=radio name="eol${i}" value=2 %s> unix<br>${CRLF}), ((defined $eol) && ($eol eq "$LF")) ? "checked" : "";

		## print jcode radio button
		##
		if ( jp::able ) {
			printf qq(<input type=radio name="jcode${i}" value=0 %s> raw${CRLF}), (! defined $code ) ? "checked" : "";
			printf qq(<input type=radio name="jcode${i}" value=1 %s> euc-jp${CRLF}), ((defined $code) && ($code eq 'euc')) ? "checked" : "";
			printf qq(<input type=radio name="jcode${i}" value=2 %s> jis${CRLF}), ((defined $code) && ($code eq 'jis')) ? "checked" : "";
			printf qq(<input type=radio name="jcode${i}" value=3 %s> shift-jis<br>${CRLF}), ((defined $code) && ($code eq 'sjis')) ? "checked" : "";;
		}

		print qq(<br><br>${CRLF});
	}

	print qq(<input type=submit value="upload">${CRLF});
	print qq(<input type=reset value="reset">${CRLF});
	print qq(</form>${CRLF});
	print_cgi_trailer;
}

sub print_webfm {
	my ( $dirname ) = @_;
	my $i;

	tmpfile::clean();

	$dirname = "" if ( ! defined $dirname );

	print_cgi_header;
	print qq(<h2>Directory /${dirname}</h2><br>${CRLF});
	&print_parentdir($dirname);
	print qq(<br><br>${CRLF});
	print qq(<form method=post action="${script_name}">${CRLF});
	print qq(<input type=hidden name="curdir" value="${dirname}">${CRLF});

	print qq(<input type=submit name="cmd" value="edit">${CRLF});
	print qq(<input type=submit name="cmd" value="upload">${CRLF});
	print qq(<input type=submit name="cmd" value="download">${CRLF});
	print qq(<input type=submit name="cmd" value="delete">${CRLF});
	print qq(<input type=submit name="cmd" value="mkdir">${CRLF});
	print qq(<input type=submit name="cmd" value="rename">${CRLF});
	print qq(<input type=submit name="cmd" value="move">${CRLF});
	print qq(<input type=submit name="cmd" value="chmod">${CRLF});
	print qq(<br><br>${CRLF});

	print qq(<table>${CRLF});
	opendir(DIR, "${TOPDIR}/${dirname}");
	$i = 1;
	foreach (sort readdir(DIR)) {
		my $file = ((defined $dirname) && ($dirname ne ""))
				? "${dirname}/$_" : $_;

		next if (/^\.+$/);
		next if (( ! -d "${TOPDIR}/${file}" ) &&
			 ( ! -f "${TOPDIR}/${file}" ));

		++$i;
		print qq(<tr><td>${CRLF});
		print qq(<input type=checkbox name="file${i}" );
		print qq(value="${file}">${CRLF});
		print qq(</td><td>${CRLF});
		if ( -d "${TOPDIR}/${file}" ) {
			print qq(<b><a href=");
			print &make_url(
					'cmd' => 'webfm',
					'file' => ${file}
					);
			print qq(">$_/</a></b>${CRLF});
		}
		elsif ( -f "${TOPDIR}/${file}" ) {
			print qq(<b>$_</b>${CRLF});
		}
		print qq(</td><td>${CRLF});
		&print_stat($file);
		print qq(</td></tr>${CRLF});
	}
	print qq(</table>${CRLF});
	print qq(</form>${CRLF});
	print_cgi_trailer;
}

############################################################################
### Utilities
############################################################################
sub parentdir {
	my ( $dir ) = @_;

	$dir =~ s/[\\\/]*[^\\\/]*$//g;
	return $dir
}

sub mode_str {
	my ( $mode ) = @_;
	my $val = "";

	my $type = ( $mode & 0170000) / 010000;
	my $per = $mode & 0777;

	if ( $type == 001 ) { $val .= "n"; }
	elsif ( $type == 002 ) { $val .= "c"; }
	elsif ( $type == 004 ) { $val .= "d"; }
	elsif ( $type == 006 ) { $val .= "b"; }
	elsif ( $type == 010 ) { $val .= "-"; }
	elsif ( $type == 012 ) { $val .= "l"; }
	elsif ( $type == 014 ) { $val .= "s"; }
	elsif ( $type == 016 ) { $val .= "w"; }
	else { $val .= "-" };

	for (my $i = 0; $i < 3; $i++) {
		$val .= ($per & 0400) ? "r" : "-";
		$val .= ($per & 0200) ? "w" : "-";
		$val .= ($per & 0100) ? "x" : "-";
		$per = $per << 3;
	}
	return $val;
}

########################################################################
### MAIN
########################################################################

jp::disable if ( ! $JP );

if (! -d ${TOPDIR}) { mkdir(${TOPDIR}, $DEFAULT_DIRMODE); };

if ($request_method =~ /^post$/i) {
	if ($content_type =~ /^application\/x-www-form-urlencoded/i) {
		my $buf;
		read(STDIN, $buf, $content_length);
		&decode_string($buf);
	}
	elsif ($content_type =~ /^multipart\/form-data/i) {
		&decode_stdin();
	}
}
elsif (( $request_method =~ /^get$/i ) &&
       ( defined $query_string )) {
	&decode_string($query_string);
}

&exec_cmd();
